/*    
       This software may be used and distributed according to the terms of
       the GNU General Public License (GPL), incorporated herein by
       reference.  Drivers based on this skeleton fall under the GPL and
       must retain the authorship (implicit copyright) notice.
 
       This program is distributed in the hope that it will be useful, but
       WITHOUT ANY WARRANTY; without even the implied warranty of
       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
       General Public License for more details.
 */
package beaconfinder;

import gnu.getopt.*;
import beaconconfig.SolverPanel;
import samplemon.*;

import cricketdaemon.clientlib.*;
import cricketdaemon.clientlib.data.*;

import javax.swing.table.*;
import java.util.*;


/**
 * An CricketDaemon application that displays statistical information
 * generated by the BeaconStatModule.
 *
 * @author Allen Miu
 */
class BeaconFinderApp 
    implements Callback, Runnable
{
    // screen size
    //static final int WIDTH  = 240;
    //static final int WIDTH  = 518;
    static final int WIDTH  = 1000;
    //    static final int HEIGHT  = 260;
    static final int HEIGHT  = 130;

    private long timeLastUpdate = 0;
    private static final long REFRESH_PERIOD = 200;

    GUI gui;
    Broker cricket;
    TableModel model;

    public static boolean useMetric    = true;
    public static boolean useBarGraph  = true;
    public static boolean useSampleMon = true;
    public static boolean scanMode     = true;

    private BarGraph bargraph;
    private SampleMonApp samplemon;

    public SampleMonApp getSampleMon() { return samplemon; }

    public BeaconFinderApp(Broker c, double stddevThresh, double maxy)
    {
	bargraph = null;
	samplemon = null;
	cricket = c;
	model = new TableModel();
	if(useBarGraph) {
	    if(useMetric)
		bargraph = new BarGraph(7, 50, WIDTH, stddevThresh);
	    else
		bargraph = new BarGraph(7, 24, WIDTH, stddevThresh);
	}

	if(useSampleMon) {
	    samplemon = new SampleMonApp(c, 5000, maxy, scanMode);
	}

	gui = new GUI("BeaconFinder", model, bargraph, samplemon, WIDTH, HEIGHT);
    }

    long prevT;
    long prevhwT;

    synchronized public void callback(CricketData data, BitSet mask)
    {
	//System.out.println("BeaconFinder::callback() "+data+" "+mask);
	if(mask.get(Broker.DEVICECOORD)) {
	    Position p = data.getDevicePosition();
	    gui.setX(Util.valueToString(p.x, 0.1));
	    gui.setY(Util.valueToString(p.y, 0.1));
	    gui.setZ(Util.valueToString(p.z, 0.1));
	}
	if(mask.get(Broker.CALCBEACONDIST)) {
	    ArrayList bd = data.getCalcBeaconDist();
	    //calcBeaconData = makeSet(calcBeaconData, bd);
	    model.setCalcBeaconData(bd.toArray());

	    /*
	      for(int i = 0; i < bd.size(); i++) {
	      BeaconRecord br = (BeaconRecord) bd.get(i);
	      System.out.println(br.uniqueName+" "+
	      Util.valueToString(Util.b2cm(br.distStat.dist),
	      0.01));
	      }
	    */
	}
	if(mask.get(Broker.BEACONSTAT) &&
	   !mask.get(Broker.SPACE) &&
	   !mask.get(Broker.DEVICECOORD)) {
	    //System.out.println("hello "+data.getBeaconsHeard().size());
	    model.setBeaconsHeard(data.getBeaconsHeard());
	    /*
	    ArrayList bd = data.getBeaconsHeard();
	    for(int i = 0; i < bd.size(); i++) {
		BeaconRecord br = (BeaconRecord) bd.get(i);
		System.out.println(br.uniqueName);
	    }
	    */

	    // data.getTimeStamp should be the hardware timestamp of the 
	    // last beacon sampled.  This is not quite correct as calls to
	    // the getTimeStamp() may not be monotonically increasing
	    // But this is a good enough approx for now.
	    //System.out.println("data.timeStamp: "+data.getHWTimeStamp()+
	    //" CurrentTime: "+System.currentTimeMillis());
	    
	    // data.getTimeStamp() == time when beacon arrives at CricketDaemon (local clock)
	    // data.getHWTimeStamp() == time when last beacon arrives in listener (hw clock)
	    //model.setClockOffset(data.getTimeStamp() - System.currentTimeMillis());
	    
	    /*
	      System.err.println("ts: "+data.getTimeStamp()+" "+
	      (data.getTimeStamp() - prevT)+" "+
	      data.getHWTimeStamp()+" "+
	      (data.getHWTimeStamp()-prevhwT));
	    */

	    prevT = data.getTimeStamp();
	    prevhwT = data.getHWTimeStamp();

	    model.setClockOffset(data.getHWTimeStamp() - data.getTimeStamp());
	    notify();
	}
    }

    private void update()
    {
	long curTime = System.currentTimeMillis();
	if(timeLastUpdate < curTime - REFRESH_PERIOD) {
	    model.updateBeaconsHeard();
	    if(bargraph != null)
		bargraph.updateGraph(model);
	    timeLastUpdate = curTime;
	}
    }

    public void run()
    {
	while(true) {
	    try {
		synchronized(this) {
		    wait(REFRESH_PERIOD);
		    update();
		}
	    } catch (InterruptedException ie) {
		
	    }
	}
    }

    public static void usage()
    {
	System.err.println("Usage: java beaconfinder.BeaconFinderApp [-H history_ms][-s stddevthresh_cm][-y max_y_axis_cm]\n");
	System.exit(1);
    }

    public static void main(String[] args) 
    {
	Getopt g = new Getopt("BeaconFinderApp", args, "H:s:Sy:");
	double stddevThresh = SolverPanel.STDDEV_THRESH;
	double history = 5000; // in ms
	double maxy    = 500; // 5 meters

	int c;
	try {
	    while((c = g.getopt()) != -1) {
		switch(c) {
		case 'H':
		    history = Double.parseDouble(g.getOptarg());
		    break;
		case 's':
		    stddevThresh = Double.parseDouble(g.getOptarg());
		    break;
		case 'S':
		    scanMode = false;
		    break;
		case 'y':
		    maxy    = Double.parseDouble(g.getOptarg());
		    break;
		default:
		    usage();
		}
	    }
	}
	catch (Exception e) {
	    usage();
	}

	Broker s = new ServerBroker();
	BeaconFinderApp bf = new BeaconFinderApp(s, stddevThresh, maxy);
	bf.getSampleMon().setHistory(history);

	BitSet mask = new BitSet();
	mask.set(Broker.BEACONSTAT);
	mask.set(Broker.DEVICECOORD);
	mask.set(Broker.CALCBEACONDIST);
	s.register(bf, mask);
	s.start();
	Thread t = new Thread(bf);
	t.start();
	try { t.join(); } catch (InterruptedException ie) { }
	s.interrupt();
    }
}
