/*    
       This software may be used and distributed according to the terms of
       the GNU General Public License (GPL), incorporated herein by
       reference.  Drivers based on this skeleton fall under the GPL and
       must retain the authorship (implicit copyright) notice.
 
       This program is distributed in the hope that it will be useful, but
       WITHOUT ANY WARRANTY; without even the implied warranty of
       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
       General Public License for more details.
 */
package cricketdaemon.space;

import cricketdaemon.Util;
import java.util.*;

/** 
 * Holds the distance distributions from a single beacon. 
 * Implements the mean, mode methods.
 *
 * @author Allen Miu, Bodhi Priyantha
 */
public class BeaconDistanceEstimator {

    protected String name;      // Beacon unique identify
    protected String space;
    protected int beaconID;
    protected Hashtable distanceDistribution;

    // caches the distance value after analysis
    private double meanDistance = -1;
    private double modeDistance = -1;

    public double compValue;

    /**
     * Constructor, Called with at least one distance value !!
     */
    BeaconDistanceEstimator(String s, int bID, int value)
    {
	distanceDistribution = new Hashtable();
	space = s;
	bID   = beaconID;
	name  = space+"-"+beaconID;
	insert(value);
    }

    /**
     * Inserts a distance value the distance to the beacon is rounded
     * acoording to PRECISION.  If the nuber of data points for a given
     * distance reaches ENOUGHPOINTS, then return value is true.  
     */
    public boolean insert(int dist)
    {
	// Tells if at least one distance has reached threshold frequency
	boolean enough_data = false;

	dist /= SpaceModule.PRECISION; // rounds down 
	dist *= SpaceModule.PRECISION; // the value

	Integer distance  = new Integer(dist);
	Object  frequency = distanceDistribution.get(distance);
	Integer newFreq   = new Integer(1);
	if(frequency != null) {  
	    // increment frequency
	    newFreq = 
		new Integer(((Integer) frequency).intValue() + 1);
	    
	    distanceDistribution.put(distance, newFreq);
	} 
	else { // insert a new distance and a frequency
	    distanceDistribution.put(distance, newFreq);
	}
	
	//count++; // adjusts total data points
	enough_data = newFreq.intValue() >= SpaceModule.ENOUGHPOINTS;

	return enough_data;
    }

    public DistanceEstimate distance(int meanMode)
    {
	double result = -1.0;
	if(meanMode == SpaceModule.MEAN)
	    result = meanDistance = mean();
	else
	    result = modeDistance = mode();
	
	if(result >= 0.0)
	    return new DistanceEstimate( space, beaconID, null, result );
	else 
	    return null;
    }
	
    /**
     * Returns the weighted mean of distance distribution.
     */
    private double mean()
    {
	Util.log("SpaceBeaconReading::mean() "+name);
	int total_dis    = 0;
	int total_sample = 0;

	// compute weighted average
	Enumeration keys = distanceDistribution.keys();
	while(keys.hasMoreElements()) {
	    Integer Dist = (Integer) keys.nextElement();
	    int dist = Dist.intValue();
	    int freq = ((Integer) distanceDistribution.get(Dist)).intValue();
	    total_dis    += freq*dist;
	    total_sample += freq;
	}
	//System.out.println(total_dis+" "+total_sample);
	if(total_sample >= SpaceModule.MEANTHRESH)
	    return (double) total_dis/total_sample;
	else
	    return -1.0;
    }

    /**
     * Returns the mode of distance distribution.
     */
    private double mode() 
    {
	Util.log("SpaceBeaconReading::mode() "+name);
	int mode_freq = 0;
	int mode_dist = 0; // running sum of modes (in case of multiple modes) 
	int mode_count= 0; // used to get the mean in case of multiple modes
	
	Enumeration keys = distanceDistribution.keys();
	while(keys.hasMoreElements()) {
	    Integer Dist = (Integer) keys.nextElement();
	    int dist = Dist.intValue();
	    int freq = ((Integer) distanceDistribution.get(Dist)).intValue();
	    if ( mode_freq < freq ) {  // a new mode
		mode_freq = freq;
		mode_dist = dist;
		mode_count = 1;
	    } 
	    else if ( mode_freq == freq) { // mutiple modes
		mode_dist += dist;
		mode_count++;
	    } // else do nothing
	}

	if (mode_freq >= SpaceModule.MODETHRESH)
	    // mode_count cannot be zero since mode_freq test
	    return (double) mode_dist/mode_count;
	else 
	    return -1.0;
    }

    public String toString() {
	return name;
    }
}

